﻿using System;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Media.Animation;
using System.Windows.Media.Media3D;

namespace Example06_3DAnimation02
{
    /// <summary>
    /// MainWindow.xaml の相互作用ロジック
    /// </summary>
    public partial class MainWindow : Window
    {
        // コンストラクター
        public MainWindow()
        {
            InitializeComponent();
            InitializeWindow();
        }

        private const double TIME = 4;         // Cube自転の周期
        private const int NUM_C = 3;           // Cubeの数
        private AnimationClock[] _clockSR = new AnimationClock[NUM_C];    // Cube自転
        private AnimationClock[] _clockVis = new AnimationClock[NUM_C];   // CubeのVisibility
        private AnimationClock _clockScY0;     // 
        private AnimationClock _clockScY1;

        // [Avi]ボタンクリックイベント
        private void btnAvi_Click(object sender, RoutedEventArgs e)
        {
            var filepath = GetAviFilePath();
            if (0 < filepath.Length)
            {
                SaveAsAvi(filepath);
                MessageBox.Show("OK");
            }
        }

        // [開始]ボタンクリックイベント
        private void btnStart_Click(object sender, RoutedEventArgs e)
        {
            for (int i = 0; i < NUM_C; i++)
            {
                _clockSR[i].Controller.Begin();
                _clockVis[i].Controller.Begin();
            }            
            _clockScY0.Controller.Begin();
            _clockScY1.Controller.Begin();
        }

        // [停止]ボタンクリックイベント
        private void btnStop_Click(object sender, RoutedEventArgs e)
        {
            for (int i = 0; i < NUM_C; i++)
            {
                _clockSR[i].Controller.Stop();
                _clockVis[i].Controller.Stop();
            }            
            _clockScY0.Controller.Stop();
            _clockScY1.Controller.Stop();
        }


        // Cube各面のGeometryModel3Dを返すメソッド
        private GeometryModel3D GModel3D(Brush brush, Point3DCollection vp3col)
        {
            var meshgm3d = new MeshGeometry3D();
            meshgm3d.Positions = vp3col;
            meshgm3d.TriangleIndices = new Int32Collection(new int[] { 0, 1, 2, 0, 2, 3 });
            meshgm3d.TextureCoordinates = new PointCollection(new Point[] { 
                new Point(0, 0), new Point(0, 1), new Point(1, 1), new Point(1, 0) });

            var geometry = new GeometryModel3D();
            geometry.Geometry = meshgm3d;
            geometry.Material = new DiffuseMaterial(brush);

            return geometry;
        }

        // CubeのVisibilityに関するObjectAnimationUsingKeyFramesを返すメソッド
        private ObjectAnimationUsingKeyFrames OaUkfVis(double time)
        {
            var frames = new ObjectAnimationUsingKeyFrames();

            var frame0 = new DiscreteObjectKeyFrame
            {
                KeyTime = KeyTime.FromTimeSpan(TimeSpan.FromSeconds(0)),
                Value = Visibility.Collapsed
            };
            frames.KeyFrames.Add(frame0);

            var frame1 = new DiscreteObjectKeyFrame
            {
                KeyTime = KeyTime.FromTimeSpan(TimeSpan.FromSeconds(time)),
                Value = Visibility.Visible
            };
            frames.KeyFrames.Add(frame1);

            return frames;
        }

        // Aviファイル保存パス取得
        private string GetAviFilePath()
        {
            var dlg = new Microsoft.Win32.SaveFileDialog();
            dlg.Filter = "aviファイル|*.avi;*.avi";
            var path = String.Empty;

            if (dlg.ShowDialog() == true)
            {
                path = dlg.FileName;
            }

            return path;
        }

        // Cubeを生成するメソッド
        private void CreateCube()
        {
            // Cube六面の形状指定
            Point3DCollection[] p3col = new Point3DCollection[6];
            p3col[0] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "1,-1,-1 1,-1,1 -1,-1,1 -1,-1,-1")));    // 底面
            p3col[1] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "-1,1,-1 -1,1,1 1,1,1 1,1,-1")));        // 上面
            p3col[2] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "-1,1,-1 -1,-1,-1 -1,-1,1 -1,1,1")));    // 左横面
            p3col[3] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "1,1,1 1,-1,1 1,-1,-1 1,1,-1")));        // 右横面
            p3col[4] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "1,1,-1 1,-1,-1 -1,-1,-1 -1,1,-1")));    // 奥
            p3col[5] = ((Point3DCollection)(new Point3DCollectionConverter().ConvertFromString(
                    "-1,1,1 -1,-1,1 1,-1,1 1,1,1")));        // 手前

            // Cube六面の色指定
            Color[] color = { Colors.Yellow, Colors.ForestGreen, Colors.MediumVioletRed,
                              Colors.DodgerBlue, Colors.Chocolate, Colors.Red};

            // 3個のCubeを生成する
            ModelUIElement3D[] cubes = new ModelUIElement3D[NUM_C];

            for (int i = 0; i < NUM_C; i++)
            {
                var groupM = new Model3DGroup();
                for (int i1 = 0; i1 < 6; i1++)
                {
                    var gmodel = GModel3D(new SolidColorBrush(color[(i * 2 + i1) % 6]), p3col[i1]);
                    groupM.Children.Add(gmodel);
                }

                cubes[i] = new ModelUIElement3D();
                cubes[i].Model = groupM;

                // CubeのTransform設定
                var groupT = new Transform3DGroup();
                cubes[i].Transform = groupT;

                // Cube自転のTransformとAnimation
                var axis3d = new AxisAngleRotation3D(new Vector3D(0, 1, 0), 0);
                groupT.Children.Add(new RotateTransform3D(axis3d));
                var da1 = new DoubleAnimation(0, 360, new Duration(TimeSpan.FromSeconds(TIME)));
                da1.RepeatBehavior = RepeatBehavior.Forever;
                _clockSR[i] = da1.CreateClock();
                axis3d.ApplyAnimationClock(AxisAngleRotation3D.AngleProperty, _clockSR[i]);
                _clockSR[i].Controller.Stop();

                // Cubeの配置
                var tr3d = new TranslateTransform3D((i - 1) * 6, 0, 0);
                groupT.Children.Add(tr3d);

                // CubeのScaleTransform3D設定(No.2 Cubeには設定しない)
                if (i < 2)
                {
                    var sc3d = new ScaleTransform3D();
                    groupT.Children.Add(sc3d);
                    
                    if (i == 0)  // No.0 Cube
                    {
                        var dasc0 = new DoubleAnimation(3,
                                      new Duration(TimeSpan.FromSeconds(TIME / 2)));
                        dasc0.AutoReverse = true;
                        dasc0.RepeatBehavior = RepeatBehavior.Forever;
                        _clockScY0 = dasc0.CreateClock();
                        sc3d.ApplyAnimationClock(ScaleTransform3D.ScaleYProperty, _clockScY0);
                        _clockScY0.Controller.Stop();
                    }
                    else         // No.1 Cube
                    {
                        var dasc1 = new DoubleAnimation(2,
                                      new Duration(TimeSpan.FromSeconds(TIME / 2)));
                        dasc1.AutoReverse = true;
                        dasc1.RepeatBehavior = RepeatBehavior.Forever;
                        _clockScY1 = dasc1.CreateClock();
                        sc3d.ApplyAnimationClock(ScaleTransform3D.ScaleYProperty, _clockScY1);
                        _clockScY1.Controller.Stop();
                    }                    
                }

                // CubeのVisibilityのAnimation設定
                cubes[i].Visibility = Visibility.Collapsed;
                if (i == 0)
                    _clockVis[i] = OaUkfVis(TIME).CreateClock();
                else if (i == 1)
                    _clockVis[i] = OaUkfVis(0).CreateClock();
                else
                    _clockVis[i] = OaUkfVis(TIME * 2).CreateClock();
                cubes[i].ApplyAnimationClock(UIElement.VisibilityProperty, _clockVis[i]);
                _clockVis[i].Controller.Stop();

                // ViewportへのCube配置
                myViewport.Children.Add(cubes[i]);
            }            
        }

        private void InitializeWindow()
        {
            CreateCube();
        }

        // Aviファイル作成
        private void SaveAsAvi(string filepath)
        {
            var fps = 48;                                // フレームレート
            var totalFrames = (int)(fps * TIME * 3);     // 全フレーム数
            var secs = Enumerable.Range(0, totalFrames).Select(t => (((double)t) / fps));
            var aviManager = new AviFile.AviManager(filepath, false);
            AviFile.VideoStream aviStream = null;
            var tempFile = "frame.png";

            foreach (var sec in secs)
            {
                for (int i = 0; i < NUM_C; i++)
                {
                    _clockSR[i].Controller.SeekAlignedToLastTick(TimeSpan.FromSeconds(sec),
                                                                 TimeSeekOrigin.BeginTime);
                    _clockVis[i].Controller.SeekAlignedToLastTick(TimeSpan.FromSeconds(sec),
                                                                  TimeSeekOrigin.BeginTime);
                }
                _clockScY0.Controller.SeekAlignedToLastTick(TimeSpan.FromSeconds(sec),
                                                            TimeSeekOrigin.BeginTime);
                _clockScY1.Controller.SeekAlignedToLastTick(TimeSpan.FromSeconds(sec),
                                                            TimeSeekOrigin.BeginTime);
                cvs1.UpdateLayout();
                cvs1.SaveImage(tempFile);
                var bmp = new System.Drawing.Bitmap(tempFile);
                if (aviStream == null)
                    aviStream = aviManager.AddVideoStream(true, fps, bmp);
                else
                    aviStream.AddFrame(bmp);
                bmp.Dispose();
            }
            aviManager.Close();

            for (int i = 0; i < NUM_C; i++)
            {
                _clockSR[i].Controller.Stop();
                _clockVis[i].Controller.Stop();
            }
            _clockScY0.Controller.Stop();
            _clockScY1.Controller.Stop();

            System.IO.File.Delete(tempFile);
        }

    }

    public static class CanvasExtension
    {
        /// <summary> [SaveImage]：Canvasの画像をファイルに保存するCanvasクラスの拡張メソッド 
        /// 出典：http://qiita.com/hugo-sb/items/894914f6bbe224a45d49
        /// </summary>
        /// <param name="canvas"></param>
        /// <param name="path">保存する画像ファイル名</param>
        /// <param name="encoder">BitmapEncoderが指定されなかった場合はPNG形式になる</param>
        public static void SaveImage(this Canvas canvas, string path, BitmapEncoder encoder = null)
        {
            // レイアウトを再計算させる
            var size = new Size(canvas.ActualWidth, canvas.ActualHeight);
            canvas.Measure(size);
            canvas.Arrange(new Rect(size));

            // VisualObjectをBitmapに変換する
            var renderBitmap = new RenderTargetBitmap((int)size.Width,       // 画像の幅
                                                      (int)size.Height,      // 画像の高さ
                                                      96.0d,                 // 横96.0DPI
                                                      96.0d,                 // 縦96.0DPI
                                                      PixelFormats.Pbgra32); // 32bit(RGBA各8bit)
            renderBitmap.Render(canvas);

            // 出力用の FileStream を作成する
            using (var stream = new System.IO.FileStream(path, System.IO.FileMode.Create))
            {
                // 変換したBitmapをエンコードしてFileStreamに保存する。
                // BitmapEncoder が指定されなかった場合は、PNG形式とする。
                encoder = encoder ?? new PngBitmapEncoder();
                encoder.Frames.Add(BitmapFrame.Create(renderBitmap));
                encoder.Save(stream);
            }
        }

    }

}
